# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import logging
import posixpath

from datetime import datetime, timezone
from typing import Optional

import attr
import attr.validators
import cattr

from pyparsing import Literal as L, Word, Optional as OptionalItem
from pyparsing import printables as _printables, rest_of_line
from pyparsing import ParseException

from linehaul.ua import UserAgent, parser as user_agents


logger = logging.getLogger(__name__)


_cattr = cattr.Converter()
_cattr.register_structure_hook(
    datetime,
    lambda d, t: datetime.strptime(d[5:-4], "%d %b %Y %H:%M:%S").replace(
        tzinfo=timezone.utc
    ),
)


class UnparseableEvent(Exception):
    pass


class _NullValue:
    pass


NullValue = _NullValue()


printables = "".join(set(_printables + " " + "\t") - {"|", "@"})

PIPE = L("|").suppress()

AT = L("@").suppress()

NULL = L("(null)")
NULL.set_parse_action(lambda s, l, t: NullValue)

TIMESTAMP = Word(printables).set_name("Timestamp")
TIMESTAMP = TIMESTAMP.set_results_name("timestamp")

COUNTRY_CODE = Word(printables).set_name("Country Code")
COUNTRY_CODE = COUNTRY_CODE.set_results_name("country_code")

URL = Word(printables).set_name("URL")
URL = URL.set_results_name("url")

REQUEST = TIMESTAMP + PIPE + OptionalItem(COUNTRY_CODE) + PIPE + URL

PROJECT_NAME = NULL | Word(printables)
PROJECT_NAME = PROJECT_NAME.set_results_name("project_name")
PROJECT_NAME.set_name("Project Name")

VERSION = NULL | Word(printables)
VERSION = VERSION.set_results_name("version")
VERSION.set_name("Version")

PACKAGE_TYPE = NULL | (
    L("sdist")
    | L("bdist_wheel")
    | L("bdist_dmg")
    | L("bdist_dumb")
    | L("bdist_egg")
    | L("bdist_msi")
    | L("bdist_rpm")
    | L("bdist_wininst")
)
PACKAGE_TYPE = PACKAGE_TYPE.set_results_name("package_type")
PACKAGE_TYPE.set_name("Package Type")

PROJECT = PROJECT_NAME + PIPE + VERSION + PIPE + PACKAGE_TYPE

TLS_PROTOCOL = NULL | Word(printables)
TLS_PROTOCOL = TLS_PROTOCOL.set_results_name("tls_protocol")
TLS_PROTOCOL.set_name("TLS Protocol")

TLS_CIPHER = NULL | Word(printables)
TLS_CIPHER = TLS_CIPHER.set_results_name("tls_cipher")
TLS_CIPHER.set_name("TLS Cipher")

TLS = TLS_PROTOCOL + PIPE + TLS_CIPHER

USER_AGENT = rest_of_line
USER_AGENT = USER_AGENT.set_results_name("user_agent")
USER_AGENT.set_name("UserAgent")

V1_HEADER = OptionalItem(L("1").suppress() + AT)

MESSAGE_v1 = V1_HEADER + REQUEST + PIPE + PROJECT + PIPE + USER_AGENT
MESSAGE_v1.leave_whitespace()

V2_HEADER = L("2").suppress() + AT

MESSAGE_v2 = V2_HEADER + REQUEST + PIPE + TLS + PIPE + PROJECT + PIPE + USER_AGENT
MESSAGE_v2.leave_whitespace()

V3_HEADER = L("download")
MESSAGE_v3 = (
    V3_HEADER + PIPE + REQUEST + PIPE + TLS + PIPE + PROJECT + PIPE + USER_AGENT
)

SIMPLE_HEADER = L("simple")
MESSAGE_SIMPLE = (
    SIMPLE_HEADER + PIPE + REQUEST + PIPE + TLS + PIPE + PIPE + PIPE + PIPE + USER_AGENT
)

MESSAGE = MESSAGE_SIMPLE | MESSAGE_v3 | MESSAGE_v2 | MESSAGE_v1


@enum.unique
class PackageType(enum.Enum):
    bdist_dmg = "bdist_dmg"
    bdist_dumb = "bdist_dumb"
    bdist_egg = "bdist_egg"
    bdist_msi = "bdist_msi"
    bdist_rpm = "bdist_rpm"
    bdist_wheel = "bdist_wheel"
    bdist_wininst = "bdist_wininst"
    sdist = "sdist"


@attr.s(slots=True, frozen=True)
class File:
    filename = attr.ib(validator=attr.validators.instance_of(str))
    project = attr.ib(validator=attr.validators.instance_of(str))
    version = attr.ib(validator=attr.validators.instance_of(str))
    type = attr.ib(type=PackageType)


@attr.s(slots=True, frozen=True)
class Download:
    timestamp = attr.ib(type=datetime)
    url = attr.ib(validator=attr.validators.instance_of(str))
    project = attr.ib(validator=attr.validators.instance_of(str))
    file = attr.ib(type=File)
    tls_protocol = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    tls_cipher = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    country_code = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    details = attr.ib(type=Optional[UserAgent], default=None)


@attr.s(slots=True, frozen=True)
class Simple:
    timestamp = attr.ib(type=datetime)
    url = attr.ib(validator=attr.validators.instance_of(str))
    project = attr.ib(validator=attr.validators.instance_of(str))
    tls_protocol = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    tls_cipher = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    country_code = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    details = attr.ib(type=Optional[UserAgent], default=None)


def _value_or_none(value):
    if value is NullValue or value == "":
        return None
    else:
        return value


def parse(message):
    try:
        parsed = MESSAGE.parse_string(message, parseAll=True)
    except ParseException as exc:
        raise UnparseableEvent("{!r} {}".format(message, exc)) from None

    data = {}
    data["timestamp"] = parsed.timestamp
    data["tls_protocol"] = _value_or_none(parsed.tls_protocol)
    data["tls_cipher"] = _value_or_none(parsed.tls_cipher)
    data["country_code"] = _value_or_none(parsed.country_code)
    data["url"] = parsed.url
    data["file"] = {}
    data["file"]["filename"] = posixpath.basename(parsed.url)
    data["file"]["project"] = _value_or_none(parsed.project_name)
    data["file"]["version"] = _value_or_none(parsed.version)
    data["file"]["type"] = _value_or_none(parsed.package_type)

    if parsed[0] == "download":
        data["project"] = _value_or_none(parsed.project_name)
        result = _cattr.structure(data, Download)
    elif parsed[0] == "simple":
        data["project"] = parsed.url.split("/")[2]
        result = _cattr.structure(data, Simple)
    else:
        result = _cattr.structure(data, Download)

    try:
        ua = user_agents.parse(parsed.user_agent)
        if ua is None:
            return  # Ignored user agents mean we'll skip trying to log this event
    except user_agents.UnknownUserAgentError:
        pass
    else:
        result = attr.evolve(result, details=ua)

    return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
import os.path

import pytest
import yaml

from hypothesis import given, strategies as st

from linehaul.events.parser import Download, UnparseableEvent, parse, _cattr


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures")


def _load_event_fixtures(fixture_dir):
    fixtures = os.listdir(fixture_dir)
    for filename in fixtures:
        with open(os.path.join(fixture_dir, filename), "r") as fp:
            fixtures = yaml.load(fp.read(), Loader=yaml.Loader)
        for fixture in fixtures:
            event = fixture.pop("event")
            result = fixture.pop("result")
            expected = (
                _cattr.structure(result, Download)
                if isinstance(result, dict)
                else result
            )
            assert fixture == {}
            yield event, expected


# @pytest.mark.parametrize(("event_data", "expected"), _load_event_fixtures(FIXTURE_DIR))
@pytest.mark.skip(reason="needs updated!")
def test_download_parsing(event_data, expected):
    if inspect.isclass(expected) and issubclass(expected, Exception):
        with pytest.raises(expected):
            parse(event_data)
    else:
        assert parse(event_data) == expected


@given(st.text(alphabet=st.characters(blacklist_characters=["\n"])))
def test_invalid_event(data):
    with pytest.raises(UnparseableEvent):
        parse(data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import re

import cattr
import packaging.version

from packaging.specifiers import SpecifierSet

from linehaul.ua.datastructures import UserAgent
from linehaul.ua.impl import ParserSet, UnableToParse, ua_parser, regex_ua_parser


logger = logging.getLogger(__name__)


class UnknownUserAgentError(ValueError):
    pass


# Note: This is a ParserSet, not a ParserList, parsers that have been registered with
#       it may be called in any order. That means that all of our parsers need to be
#       ordering independent.
_parser = ParserSet()


@_parser.register
@ua_parser
def Pip6UserAgent(user_agent):
    # We're only concerned about pip user agents.
    if not user_agent.startswith("pip/"):
        raise UnableToParse

    # This format was brand new in pip 6.0, so we'll need to restrict it
    # to only versions of pip newer than that.
    version_str = user_agent.split()[0].split("/", 1)[1]
    version = packaging.version.parse(version_str)
    if version not in SpecifierSet(">=6", prereleases=True):
        raise UnableToParse

    try:
        return json.loads(user_agent.split(maxsplit=1)[1])
    except (json.JSONDecodeError, UnicodeDecodeError, IndexError):
        raise UnableToParse from None


@_parser.register
@regex_ua_parser(
    (
        r"^pip/(?P<version>\S+) (?P<impl_name>\S+)/(?P<impl_version>\S+) "
        r"(?P<system_name>\S+)/(?P<system_release>\S+)$"
    )
)
def Pip1_4UserAgent(*, version, impl_name, impl_version, system_name, system_release):
    # This format was brand new in pip 1.4, and went away in pip 6.0, so
    # we'll need to restrict it to only versions of pip between 1.4 and 6.0.
    if version not in SpecifierSet(">=1.4,<6", prereleases=True):
        raise UnableToParse

    data = {"installer": {"name": "pip", "version": version}}

    if impl_name.lower() != "unknown":
        data.setdefault("implementation", {})["name"] = impl_name

    if impl_version.lower() != "unknown":
        data.setdefault("implementation", {})["version"] = impl_version

    if system_name.lower() != "unknown":
        data.setdefault("system", {})["name"] = system_name

    if system_release.lower() != "unknown":
        data.setdefault("system", {})["release"] = system_release

    if impl_name.lower() == "cpython":
        data["python"] = impl_version

    return data


@_parser.register
@regex_ua_parser(r"^Python-urllib/(?P<python>\d\.\d) distribute/(?P<version>\S+)$")
def DistributeUserAgent(*, python, version):
    return {"installer": {"name": "distribute", "version": version}, "python": python}


@_parser.register
@regex_ua_parser(
    r"^Python-urllib/(?P<python>\d\.\d) setuptools/(?P<version>\S+)$",
    r"^setuptools/(?P<version>\S+) Python-urllib/(?P<python>\d\.\d)$",
)
def SetuptoolsUserAgent(*, python, version):
    return {"installer": {"name": "setuptools", "version": version}, "python": python}


@_parser.register
@regex_ua_parser(r"pex/(?P<version>\S+)$")
def PexUserAgent(*, version):
    return {"installer": {"name": "pex", "version": version}}


@_parser.register
@regex_ua_parser(r"^conda/(?P<version>\S+)(?: .+)?$")
def CondaUserAgent(*, version):
    return {"installer": {"name": "conda", "version": version}}


@_parser.register
@regex_ua_parser(r"^Bazel/(?:release\s+)?(?P<version>.+)$")
def BazelUserAgent(*, version):
    return {"installer": {"name": "Bazel", "version": version}}


@_parser.register
@regex_ua_parser(r"^bandersnatch/(?P<version>\S+) \(.+\)$")
def BandersnatchUserAgent(*, version):
    return {"installer": {"name": "bandersnatch", "version": version}}


@_parser.register
@regex_ua_parser(r"devpi-server/(?P<version>\S+) \(.+\)$")
def DevPIUserAgent(*, version):
    return {"installer": {"name": "devpi", "version": version}}


@_parser.register
@regex_ua_parser(r"^z3c\.pypimirror/(?P<version>\S+)$")
def Z3CPyPIMirrorUserAgent(*, version):
    return {"installer": {"name": "z3c.pypimirror", "version": version}}


@_parser.register
@regex_ua_parser(r"^Artifactory/(?P<version>\S+)$")
def ArtifactoryUserAgent(*, version):
    return {"installer": {"name": "Artifactory", "version": version}}


@_parser.register
@regex_ua_parser(r"^Nexus/(?P<version>\S+)")
def NexusUserAgent(*, version):
    return {"installer": {"name": "Nexus", "version": version}}


@_parser.register
@regex_ua_parser(r"^pep381client(?:-proxy)?/(?P<version>\S+)$")
def PEP381ClientUserAgent(*, version):
    return {"installer": {"name": "pep381client", "version": version}}


@_parser.register
@regex_ua_parser(r"^maturin/(?P<version>\S+)$")
def MaturinUserAgent(*, version):
    return {"installer": {"name": "maturin", "version": version}}


@_parser.register
@regex_ua_parser(r"^pdm/(?P<version>\S+) (?P<impl_name>\S+)/(?P<impl_version>\S+)$")
def PDMUserAgent(*, version, impl_name, impl_version):
    return {
        "installer": {"name": "pdm", "version": version},
        "implementation": {"name": impl_name, "version": impl_version},
    }


@_parser.register
@regex_ua_parser(
    r"^poetry/(?P<version>\S+) (?P<impl_name>\S+)/(?P<impl_version>\S+) "
    r"(?P<system_name>\S+)/(?P<system_release>\S+)?$"
)
def PoetryUserAgent(*, version, impl_name, impl_version, system_name, system_release):
    return {
        "installer": {"name": "poetry", "version": version},
        "implementation": {"name": impl_name, "version": impl_version},
        "system": {"name": system_name, "release": system_release},
    }


@_parser.register
@regex_ua_parser(
    r"^twine/(?P<version>\S+)(?: .+)? (?P<impl_name>\S+)/(?P<impl_version>\S+)$"
)
def TwineUserAgent(*, version, impl_name, impl_version):
    return {
        "installer": {"name": "twine", "version": version},
        "implementation": {"name": impl_name, "version": impl_version},
    }


@_parser.register
@ua_parser
def UvUserAgent(user_agent):
    # We're only concerned about uv user agents.
    if not user_agent.startswith("uv/"):
        raise UnableToParse

    # This format was brand new in uv 0.1.22, so we'll need to restrict it
    # to only versions of uv newer than that.
    version_str = user_agent.split()[0].split("/", 1)[1]
    version = packaging.version.parse(version_str)
    if version not in SpecifierSet(">=0.1.22", prereleases=True):
        raise UnableToParse

    try:
        return json.loads(user_agent.split(maxsplit=1)[1])
    except (json.JSONDecodeError, UnicodeDecodeError, IndexError):
        raise UnableToParse from None


# TODO: We should probably consider not parsing this specially, and moving it to
#       just the same as we treat browsers, since we don't really know anything
#       about it-- including whether or not the version of Python mentioned is
#       the one they're going to install it into or not. The one real sticking
#       point is that before pip 1.4, pip just used the default urllib2 UA, so
#       right now we're counting pip 1.4 in here... but pip 1.4 usage is probably
#       low enough not to worry about that any more.
@_parser.register
@regex_ua_parser(r"^Python-urllib/(?P<python>\d\.\d)$")
def URLLib2UserAgent(*, python):
    return {"python": python}


# TODO: We should probably consider not parsing this specially, and moving it to
#       just the same as we treat browsers, since we don't really know anything
#       about it and the version of requests isn't very useful in general.
@_parser.register
@regex_ua_parser(r"^python-requests/(?P<version>\S+)(?: .+)?$")
def RequestsUserAgent(*, version):
    return {"installer": {"name": "requests", "version": version}}


@_parser.register
@regex_ua_parser(
    (
        r"^Homebrew/(?P<version>\S+) "
        r"\(Macintosh; Intel (?:Mac OS X|macOS) (?P<osx_version>[^)]+)\)(?: .+)?$"
    )
)
def HomebrewUserAgent(*, version, osx_version):
    return {
        "installer": {"name": "Homebrew", "version": version},
        "distro": {"name": "OS X", "version": osx_version},
    }


# TODO: It would be nice to maybe break more of these apart to try and get more insight
#       into the OSs that people are installing packages into (similiar to Homebrew).
@_parser.register
@regex_ua_parser(
    re.compile(
        r"""
    (?:
        ^fetch\ libfetch/\S+$ |
        ^libfetch/\S+$ |
        ^OpenBSD\ ftp$ |
        ^MacPorts/? |
        ^NetBSD-ftp/ |
        ^slapt-get |
        ^pypi-install/ |
        ^slackrepo$ |
        ^PTXdist |
        ^GARstow/ |
        ^xbps/
    )
    """,
        re.VERBOSE,
    )
)
def OSUserAgent():
    return {"installer": {"name": "OS"}}


@_parser.register
@regex_ua_parser(
    re.compile(
        r"""
        ^
        (?:
            Mozilla |
            Safari |
            wget |
            curl |
            Opera |
            aria2 |
            AndroidDownloadManager |
            com\.apple\.WebKit\.Networking/ |
            FDM\ \S+ |
            URL/Emacs |
            Firefox/ |
            UCWEB |
            Links |
            ^okhttp |
            ^Apache-HttpClient
        )
        (?:/|$)
    """,
        re.IGNORECASE | re.VERBOSE,
    )
)
def BrowserUserAgent():
    return {"installer": {"name": "Browser"}}


# TODO: It would be kind of nice to implement this as just another parser, that returns
#       None instead of a dictionary. However given that there is no inherent ordering
#       in a ParserSet, and we want this to always go last (just incase an ignore
#       pattern is overlly broad) we can't do that. It would be nice to make it possible
#       to register a parser with an explicit location in the parser set.
_ignore_re = re.compile(
    r"""
    (?:
        ^Datadog\ Agent/ |
        ^\(null\)$ |
        ^WordPress/ |
        ^Chef\ (?:Client|Knife)/ |
        ^Ruby$ |
        ^Slackbot-LinkExpanding |
        ^TextualInlineMedia/ |
        ^WeeChat/ |
        ^Download\ Master$ |
        ^Java/ |
        ^Go\ \d\.\d\ package\ http$ |
        ^Go-http-client/ |
        ^GNU\ Guile$ |
        ^github-olee$ |
        ^YisouSpider$ |
        ^Apache\ Ant/ |
        ^Salt/ |
        ^ansible-httpget$ |
        ^ltx71\ -\ \(http://ltx71.com/\) |
        ^Scrapy/ |
        ^spectool/ |
        Nutch |
        ^AWSBrewLinkChecker/ |
        ^Y!J-ASR/ |
        ^NSIS_Inetc\ \(Mozilla\)$ |
        ^Debian\ uscan |
        ^Pingdom\.com_bot_version_\d+\.\d+_\(https?://www.pingdom.com/\)$ |
        ^MauiBot\ \(crawler\.feedback\+dc@gmail\.com\)$
    )
    """,
    re.VERBOSE,
)


def parse(user_agent: str) -> UserAgent | None:
    try:
        return cattr.structure(_parser(user_agent), UserAgent)
    except UnableToParse:
        # If we were not able to parse the user agent, then we have two options, we can
        # either raise an `UnknownUserAgentError` or we can return None to explicitly
        # say that we opted not to parse this user agent. To determine which option we
        # pick we'll match against a regex of UAs to ignore, if we match then we'll
        # return a None to indicate to our caller that we couldn't parse this UA, but
        # that it was an expected inability to parse. Otherwise we'll raise an
        # `UnknownUserAgentError` to indicate that it as an unexpected inability to
        # parse.
        if _ignore_re.search(user_agent) is not None:
            return None

        raise UnknownUserAgentError from None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections
import logging
import random
import re


logger = logging.getLogger(__name__)


class UnableToParse(Exception):
    pass


class UserAgentParser(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def name(self):
        """
        Returns the name of this parser, useful for things like logging etc.
        """

    @abc.abstractmethod
    def __call__(self, ua):
        """
        Actually parses the user agent, returning a dictionary containing all of the
        relevant parsed information. If this method is unable to parse the user agent
        then it can raise a ``UnableToParse`` exception to indicate that it can't parse
        the given UA.
        """


class CallbackUserAgentParser(UserAgentParser):
    def __init__(self, callback, *, name=None):
        if name is None:
            name = callback.__name__

        self._callback = callback
        self._name = name

    @property
    def name(self):
        return self._name

    def __call__(self, ua):
        return self._callback(ua)


def ua_parser(fn):
    return CallbackUserAgentParser(fn)


class RegexUserAgentParser(UserAgentParser):
    def __init__(self, regexes, handler, *, name=None):
        if name is None:
            name = handler.__name__

        self._regexes = [
            re.compile(regex) if isinstance(regex, str) else regex for regex in regexes
        ]
        self._handler = handler
        self._name = name

    @property
    def name(self):
        return self._name

    def __call__(self, user_agent):
        for regex in self._regexes:
            matched = regex.search(user_agent)

            # If we've matched this particuar regex, then we'll break the loop here and
            # go onto finishing parsing.
            if matched is not None:
                break
        else:
            # None of our regexes matched.
            raise UnableToParse

        # We need to build up the args, and kwargs of our function, we call any unnamed
        # group an arg, and pass them in, in order, and we call any named group a kwarg
        # and we pass them in by name.
        group_to_name = {v: k for k, v in matched.re.groupindex.items()}
        args, kwargs = [], {}
        for i, value in enumerate(matched.groups(), start=1):
            name = group_to_name.get(i)
            if name is not None:
                kwargs[name] = value
            else:
                args.append(value)

        # Finally, we'll call our handler with our parsed arguments, and return whatever
        # result it gives us.
        return self._handler(*args, **kwargs)


def regex_ua_parser(*regexes):
    def deco(fn):
        return RegexUserAgentParser(regexes, fn)

    return deco


class ParserSet:
    def __init__(self):
        self._parsers = []

        self._optimize_every = 1000000
        # Set the first optimize in to a reduced amount to get some basic optimization
        # done early.
        self._optimize_in = self._optimize_every * 0.25
        self._counts = collections.Counter()

    def register(self, parser, *, _randomize=True):
        self._parsers.append(parser)

        # The use of random.shuffle here is a bit quirkly, it doesn't actually help us
        # at runtime in any way. What it *does* do, is make it more likely that any
        # ordering dependence in registered parsers shows up as test failures instead
        # of being hard to find bugs in production.
        # This does make registering a parser more heavy-weight than recorded (through
        # minorly so), but this shouldn't matter since in our usage registerin is only
        # done at the module level anyways.
        if _randomize:
            random.shuffle(self._parsers)

        return parser

    def _optimize(self):
        # We're going to sort our list in place, using the value of how many times
        # a parser function has been used as the parser for a user agent to put the
        # most commonly used parsed first.
        self._parsers.sort(key=lambda p: self._counts[p], reverse=True)

        # Reduce our recorded counts just to keep the size of our counts in checks.
        # This will also implicitly act as a decay so that historical data is less
        # relevant than new data.
        self._counts.subtract({k: int(v * 0.5) for k, v in self._counts.items()})

        # Reset our marker
        self._optimize_in = self._optimize_every

    def __call__(self, user_agent):
        # Decrement our counter for how long until we will implicitly call optimize
        # on our ParserSet, and check to see if it's time to optimize or not.
        self._optimize_in -= 1
        if self._optimize_in <= 0:
            self._optimize()

        # Actually go through our registered parsers and try to use them to parse.
        for parser in self._parsers:
            try:
                parsed = parser(user_agent)

                # Record a "hit" for this parser.
                self._counts[parser] += 1

                return parsed
            except UnableToParse:
                pass
            except Exception:
                logger.error(
                    "Error parsing %r as a %s.", user_agent, parser.name, exc_info=True
                )

        raise UnableToParse



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import attr


@attr.s(slots=True, frozen=True)
class Installer:
    name = attr.ib(type=Optional[str], default=None)
    version = attr.ib(type=Optional[str], default=None)


@attr.s(slots=True, frozen=True)
class Implementation:
    name = attr.ib(type=Optional[str], default=None)
    version = attr.ib(type=Optional[str], default=None)


@attr.s(slots=True, frozen=True)
class LibC:
    lib = attr.ib(type=Optional[str], default=None)
    version = attr.ib(type=Optional[str], default=None)


@attr.s(slots=True, frozen=True)
class Distro:
    name = attr.ib(type=Optional[str], default=None)
    version = attr.ib(type=Optional[str], default=None)
    id = attr.ib(type=Optional[str], default=None)
    libc = attr.ib(type=Optional[LibC], default=None)


@attr.s(slots=True, frozen=True)
class System:
    name = attr.ib(type=Optional[str], default=None)
    release = attr.ib(type=Optional[str], default=None)


@attr.s(slots=True, frozen=True)
class UserAgent:
    installer = attr.ib(type=Optional[Installer], default=None)
    python = attr.ib(type=Optional[str], default=None)
    implementation = attr.ib(type=Optional[Implementation], default=None)
    distro = attr.ib(type=Optional[Distro], default=None)
    system = attr.ib(type=Optional[System], default=None)
    cpu = attr.ib(type=Optional[str], default=None)
    openssl_version = attr.ib(type=Optional[str], default=None)
    setuptools_version = attr.ib(type=Optional[str], default=None)
    rustc_version = attr.ib(type=Optional[str], default=None)
    ci = attr.ib(type=Optional[bool], default=None)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os.path

import cattr
import pytest
import yaml

from hypothesis import given, strategies as st

from linehaul.ua import parser
from linehaul.ua.datastructures import UserAgent

from ...strategies import version as st_version


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures")


def _load_ua_fixtures(fixture_dir):
    fixtures = os.listdir(fixture_dir)
    for filename in fixtures:
        with open(os.path.join(fixture_dir, filename), "r") as fp:
            fixtures = yaml.safe_load(fp.read())
        for fixture in fixtures:
            ua = fixture.pop("ua")
            result = fixture.pop("result")
            expected = (
                cattr.structure(result, UserAgent)
                if isinstance(result, dict)
                else result
            )
            assert fixture == {}
            yield ua, expected


@pytest.mark.parametrize(("ua", "expected"), _load_ua_fixtures(FIXTURE_DIR))
def test_user_agent_parsing(ua, expected):
    assert parser.parse(ua) == expected


def _is_valid_json(item):
    try:
        json.loads(item)
    except Exception:
        return False

    return True


class TestPip6UserAgent:
    @given(st.text().filter(lambda i: not i.startswith("pip/")))
    def test_not_pip(self, ua):
        with pytest.raises(parser.UnableToParse):
            parser.Pip6UserAgent(ua)

    @given(st_version(max_version="6"))
    def test_invalid_version(self, version):
        with pytest.raises(parser.UnableToParse):
            parser.Pip6UserAgent(f"pip/{version}")

    @given(st.text(max_size=100).filter(lambda i: not _is_valid_json(i)))
    def test_invalid_json(self, json_blob):
        with pytest.raises(parser.UnableToParse):
            parser.Pip6UserAgent(f"pip/18.0 {json_blob}")


class TestPip1_4UserAgent:
    @given(st.text().filter(lambda i: not i.startswith("pip/")))
    def test_not_pip(self, ua):
        with pytest.raises(parser.UnableToParse):
            parser.Pip1_4UserAgent(ua)

    @given(st_version(max_version="1.3") | st_version(min_version="6"))
    def test_invalid_version(self, version):
        with pytest.raises(parser.UnableToParse):
            parser.Pip1_4UserAgent(f"pip/{version} Unknown/Unknown Unknown/Unknown")

    @given(st_version(min_version="1.4", max_version="5"))
    def test_no_other_data(self, version):
        with pytest.raises(parser.UnableToParse):
            parser.Pip1_4UserAgent(f"pip/{version}")

    @given(
        st_version(min_version="1.4", max_version="5"),
        (
            st.just("Unknown")
            | st.just("Cpython")
            | st.just("PyPy")
            | st.text(
                min_size=1,
                alphabet=st.characters(
                    blacklist_categories=["Cc", "Z"], blacklist_characters="/"
                ),
            )
        ),
        (
            st.just("Unknown")
            | st_version()
            | st.text(
                min_size=1,
                alphabet=st.characters(
                    blacklist_categories=["Cc", "Z"], blacklist_characters="/"
                ),
            )
        ),
        (
            st.just("Unknown")
            | st.just("Darwin")
            | st.just("Linux")
            | st.just("Windows")
            | st.text(
                min_size=1,
                alphabet=st.characters(
                    blacklist_categories=["Cc", "Z"], blacklist_characters="/"
                ),
            )
        ),
        (
            st.just("Unknown")
            | st.just("17.7.0")
            | st.just("NT")
            | st_version()
            | st.text(
                min_size=1,
                alphabet=st.characters(
                    blacklist_categories=["Cc", "Z"], blacklist_characters="/"
                ),
            )
        ),
    )
    def test_valid_data(
        self, version, impl_name, impl_version, system_name, system_release
    ):
        ua = f"pip/{version} {impl_name}/{impl_version} {system_name}/{system_release}"

        expected = {"installer": {"name": "pip", "version": version}}
        if impl_name.lower() != "unknown":
            expected.setdefault("implementation", {})["name"] = impl_name
        if impl_version.lower() != "unknown":
            expected.setdefault("implementation", {})["version"] = impl_version
        if system_name.lower() != "unknown":
            expected.setdefault("system", {})["name"] = system_name
        if system_release.lower() != "unknown":
            expected.setdefault("system", {})["release"] = system_release
        if impl_name.lower() == "cpython":
            expected["python"] = impl_version

        assert parser.Pip1_4UserAgent(ua) == expected


class TestUvUserAgent:
    @given(st.text().filter(lambda i: not i.startswith("uv/")))
    def test_not_uv(self, ua):
        with pytest.raises(parser.UnableToParse):
            parser.UvUserAgent(ua)

    @given(st_version(max_version="0.1.21"))
    def test_invalid_version(self, version):
        with pytest.raises(parser.UnableToParse):
            parser.UvUserAgent(f"""uv/{version} {{"installer":{{"name":"uv","version":"{version}"}}}}""")

    @given(st.text(max_size=100).filter(lambda i: not _is_valid_json(i)))
    def test_invalid_json(self, json_blob):
        with pytest.raises(parser.UnableToParse):
            parser.UvUserAgent(f"uv/0.1.22 {json_blob}")

class TestParse:
    @given(st.text())
    def test_unknown_user_agent(self, user_agent):
        with pytest.raises(parser.UnknownUserAgentError):
            parser.parse(user_agent)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from hypothesis import assume, given, strategies as st, reproduce_failure, seed

from .strategies import version as st_version


class TestVersionStrategy:
    @staticmethod
    def _ver_2_list(version):
        version = [int(i) for i in version.split(".")]
        return list(
            reversed(list(itertools.dropwhile(lambda i: i == 0, reversed(version))))
        )

    @given(
        st.data(),
        st.tuples(
            st.integers(min_value=1), st.integers(min_value=1), st.integers(min_value=1)
        ).map(lambda i: ".".join(str(j) for j in i)),
    )
    def test_greater_than_minimum(self, data, min_version):
        version = data.draw(st_version(min_version=min_version))
        assert self._ver_2_list(version) >= self._ver_2_list(min_version)

    @given(
        st.data(),
        st.tuples(
            st.integers(min_value=10),
            st.integers(min_value=1),
            st.integers(min_value=1),
        ).map(lambda i: ".".join(str(j) for j in i)),
    )
    def test_less_than_maximum(self, data, max_version):
        version = data.draw(st_version(max_version=max_version))
        assert self._ver_2_list(version) <= self._ver_2_list(max_version)

    @given(
        st.data(),
        st.tuples(
            st.tuples(
                st.integers(min_value=1),
                st.integers(min_value=1),
                st.integers(min_value=1),
            ),
            st.tuples(
                st.integers(min_value=1),
                st.integers(min_value=1),
                st.integers(min_value=1),
            ),
        ).map(lambda inp: [".".join(str(i) for i in p) for p in sorted(inp)]),
    )
    def test_inbetween_min_and_max(self, data, versions):
        min_version, max_version = versions
        version = data.draw(
            st_version(min_version=min_version, max_version=max_version)
        )
        assert (
            self._ver_2_list(min_version)
            <= self._ver_2_list(version)
            <= self._ver_2_list(max_version)
        )

    @given(st.data(), st.integers(min_value=1, max_value=100))
    def test_produces_with_more_digits_than_min(self, data, min_digits):
        version = data.draw(st_version(min_digits=min_digits))
        assert len(version.split(".")) >= min_digits

    @given(st.data(), st.integers(min_value=2, max_value=100))
    def test_produces_with_less_digits_than_max(self, data, max_digits):
        version = data.draw(st_version(max_digits=max_digits))
        assert len(version.split(".")) <= max_digits

    @given(
        st.data(),
        st.tuples(
            st.integers(min_value=1, max_value=100),
            st.integers(min_value=1, max_value=100),
        ).map(lambda inp: sorted(inp)),
    )
    def test_produces_inbetween_min_and_max_digits(self, data, digits):
        min_digits, max_digits = digits
        version = data.draw(st_version(min_digits=min_digits, max_digits=max_digits))
        assert min_digits <= len(version.split(".")) <= max_digits

    @given(
        st.data(),
        st.tuples(
            st.tuples(
                st.integers(min_value=1),
                st.integers(min_value=1),
                st.integers(min_value=1),
            ),
            st.tuples(
                st.integers(min_value=1),
                st.integers(min_value=1),
                st.integers(min_value=1),
            ),
        ).map(lambda inp: [".".join(str(i) for i in p) for p in sorted(inp)]),
        st.tuples(
            st.integers(min_value=1, max_value=100),
            st.integers(min_value=1, max_value=100),
        ).map(lambda inp: sorted(inp)),
    )
    def test_mixture(self, data, versions, digits):
        min_version, max_version = versions
        min_digits, max_digits = digits

        # Check that the our minimum version doesn't have too many digits.
        # TODO: Can we remove these assumptions?
        assume(len(min_version.split(".")) <= max_digits)
        assume(len(max_version.split(".")) <= max_digits)

        version = data.draw(
            st_version(
                min_digits=min_digits,
                max_digits=max_digits,
                min_version=min_version,
                max_version=max_version,
            )
        )

        assert (
            self._ver_2_list(min_version)
            <= self._ver_2_list(version)
            <= self._ver_2_list(max_version)
        )
        assert min_digits <= len(version.split(".")) <= max_digits



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import re

import pytest

from linehaul.ua import impl


class TestDecorators:
    def test_ua_parser(self, monkeypatch):
        class FakeCallbackParser:
            def __init__(self, *args, **kwargs):
                self.args = list(args)
                self.kwargs = kwargs

        monkeypatch.setattr(impl, "CallbackUserAgentParser", FakeCallbackParser)

        def MyParserFn():
            pass

        MyParser = impl.ua_parser(MyParserFn)

        assert isinstance(MyParser, FakeCallbackParser)
        assert MyParser.args == [MyParserFn]
        assert MyParser.kwargs == {}

    @pytest.mark.parametrize("regexes", [(r"^one$",), (r"^one$", r"^two$")])
    def test_regex_us_parser(self, monkeypatch, regexes):
        class FakeRegexParser:
            def __init__(self, *args, **kwargs):
                self.args = list(args)
                self.kwargs = kwargs

        monkeypatch.setattr(impl, "RegexUserAgentParser", FakeRegexParser)

        def MyHandlerFn():
            pass

        MyParser = impl.regex_ua_parser(*regexes)(MyHandlerFn)

        assert isinstance(MyParser, FakeRegexParser)
        assert MyParser.args == [regexes, MyHandlerFn]
        assert MyParser.kwargs == {}


class TestCallbackUserAgentParser:
    def test_undefined_name(self):
        def FakeUserAgent():
            pass

        parser = impl.CallbackUserAgentParser(FakeUserAgent)
        assert parser.name == "FakeUserAgent"

    def test_explicit_name(self):
        parser = impl.CallbackUserAgentParser(lambda i: i, name="MyName")
        assert parser.name == "MyName"

    def test_returns_on_success(self):
        result = object()
        parser = impl.CallbackUserAgentParser(lambda inp: result)
        assert parser("any input") is result

    def test_passed_input(self):
        parser = impl.CallbackUserAgentParser(lambda inp: {"input": inp})
        assert parser("another input") == {"input": "another input"}


class TestRegexUserAgentParser:
    def test_undefined_name(self):
        def FakeUserAgent():
            pass

        parser = impl.RegexUserAgentParser([], FakeUserAgent)
        assert parser.name == "FakeUserAgent"

    def test_explicit_name(self):
        parser = impl.RegexUserAgentParser([], lambda i: i, name="MyName")
        assert parser.name == "MyName"

    @pytest.mark.parametrize(
        ("regexes", "input"),
        [
            ([r"^Foo Bar$"], "Foo Bar"),
            ([re.compile(r"^Foo Bar$")], "Foo Bar"),
            ([r"^Bar Foo$", re.compile(r"^Foo Bar$")], "Foo Bar"),
            ([r"^Bar Foo$", re.compile(r"^Foo Bar$")], "Bar Foo"),
        ],
    )
    def test_valid(self, regexes, input):
        result = object()
        parser = impl.RegexUserAgentParser(regexes, lambda: result)
        assert parser(input) is result

    @pytest.mark.parametrize(
        ("regexes", "input"),
        [
            ([], "literally anything"),
            ([r"^A Test String$"], "totally not a test string"),
            ([r"^One$", re.compile(r"^Two$")], "Three"),
        ],
    )
    def test_invalid(self, regexes, input):
        parser = impl.RegexUserAgentParser(regexes, None, name="AName")
        with pytest.raises(impl.UnableToParse):
            parser(input)

    def test_positional_captures(self):
        def handler(*args):
            return list(args)

        parser = impl.RegexUserAgentParser([r"^Foo (.+)$"], handler)
        assert parser("Foo Bar") == ["Bar"]

    def test_named_captures(self):
        def handler(**kwargs):
            return kwargs

        parser = impl.RegexUserAgentParser([r"^Foo (?P<thing>.+)$"], handler)
        assert parser("Foo Bar") == {"thing": "Bar"}

    def test_mixed_captures(self):
        def handler(*args, **kwargs):
            return list(args), kwargs

        parser = impl.RegexUserAgentParser(
            [r"^(\S+) (?P<thing>\S+) (\S+) (?P<another>\S+)$"], handler
        )
        assert parser("Foo Bar Widget Frob") == (
            ["Foo", "Widget"],
            {"thing": "Bar", "another": "Frob"},
        )


class TestParserSet:
    def test_valid(self):
        def raiser(inp):
            raise impl.UnableToParse

        parser = impl.ParserSet()

        parser.register(raiser)
        parser.register(lambda i: {"parsed": "data"})
        parser.register(raiser)

        assert parser("anything") == {"parsed": "data"}

    def test_cannot_parse(self):
        def raiser(inp):
            raise impl.UnableToParse

        parser = impl.ParserSet()
        parser.register(raiser)

        with pytest.raises(impl.UnableToParse):
            parser("anything")

    def test_error_while_parsing(self, caplog):
        def raiser(inp):
            raise ValueError("Oh No")

        raiser.name = "OhNoName"

        parser = impl.ParserSet()
        parser.register(raiser)

        with pytest.raises(impl.UnableToParse):
            parser("anything")

        assert caplog.record_tuples == [
            (
                "linehaul.ua.impl",
                logging.ERROR,
                "Error parsing 'anything' as a OhNoName.",
            )
        ]

    def test_optimizing(self):
        parser = impl.ParserSet()

        # Override some values to make it easier to test.
        parser._optimize_every = 100
        parser._optimize_in = 100

        def parser1(inp):
            if inp != "one":
                raise impl.UnableToParse

        def parser2(inp):
            if inp != "two" and inp != "three":
                raise impl.UnableToParse

        def parser3(inp):
            if inp != "four":
                raise impl.UnableToParse

        # We explicitly register these with the private, randomize kwarg set to
        # False, that will ensure that our default order is the ordered we registered
        # these in, which will make this test easier to write.
        parser.register(parser1, _randomize=False)
        parser.register(parser2, _randomize=False)
        parser.register(parser3, _randomize=False)

        # Check our start state makes sense.
        assert parser._parsers == [parser1, parser2, parser3]

        # Run through our parser with some input, watching the state of our optimize
        # markers as we go along.
        for i, value in enumerate(itertools.cycle(["one", "two", "three", "four"])):
            assert parser._optimize_in == parser._optimize_every - i
            parser(value)
            assert parser._optimize_in == parser._optimize_every - i - 1
            assert parser._parsers == [parser1, parser2, parser3]

            # Break out of our loop right before the optimize method would be called.
            if i + 1 >= 99:
                break

        # We should know the exact state of our counters at this point.
        assert parser._counts == {parser1: 25, parser2: 50, parser3: 24}

        # Call our parser one more time, we explictly call it with "one" here, because
        # that should ensure that the first parser has been called at least once more
        # than the third person, and the second parser should have been called almost
        # twice as many times as either.
        parser("one")

        assert parser._optimize_in == 100
        assert parser._parsers == [parser2, parser1, parser3]
        assert parser._counts == {parser1: 14, parser2: 25, parser3: 12}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hypothesis import strategies as st


INF = float("inf")


@st.composite
def line_delimited_data(draw, max_line_size, min_lines=1):
    n = draw(max_line_size)
    data = st.binary(min_size=1, max_size=n).filter(lambda d: b"\n" not in d)
    lines = draw(
        st.lists(data, min_size=min_lines).filter(
            lambda l: sum(map(len, l)) + len(l) <= n
        )
    )
    return b"\n".join(lines) + b"\n"


@st.composite
def chunked(draw, source):
    data = draw(source)

    chunk_sizes = [0]
    chunk_sizes += draw(
        st.lists(st.integers(0, len(data) - 1), unique=True).map(sorted)
    )
    chunk_sizes += [len(data)]

    return [data[u:v] for u, v in zip(chunk_sizes, chunk_sizes[1:])]


def _none_for_inf(v):
    if v is INF:
        return None
    return v


@st.composite
def version(draw, min_digits=1, max_digits=None, min_version=None, max_version=None):
    min_version_digits = None if min_version is None else len(min_version.split("."))
    max_version_digits = None if max_version is None else len(max_version.split("."))

    if min_digits < 1:
        raise ValueError("Minimum digits must be >= 1")
    if max_digits is None:
        # To determine our maximum number of digits, we're going to take the larger of
        # our default of 10 greater than the minimum, or the number of digits in the min
        # and max versions.
        max_digits = max(
            filter(None, [min_digits + 10, min_version_digits, max_version_digits])
        )
    if min_digits > max_digits:
        raise ValueError("Maximum digits must be greater than the minimum digits.")
    if min_version_digits is not None and min_version_digits > max_digits:
        raise ValueError(
            "Cannot have a minimum version with more digits than the maximum number "
            "of digits."
        )
    if max_version_digits is not None and max_version_digits > max_digits:
        raise ValueError(
            "Cannot have a maximum version with more digits than the maximum number "
            "of digits."
        )

    num_digits = draw(st.integers(min_value=min_digits, max_value=max_digits))

    if min_version is not None:
        min_version = [int(i) for i in min_version.split(".")]
    else:
        min_version = [0]

    # We need to pad out the minimum version so that it matches our number of digits.
    min_version += [0 for _ in range(num_digits - len(min_version))]

    if max_version is not None:
        # If we were given a max range, than we want to pad it out to zeros to match
        # the number of digits we're trying to generate.
        max_version = [int(i) for i in max_version.split(".")]
        max_version += [0 for _ in range(num_digits - len(max_version))]
    else:
        # If we were not given a max range, we want to have an infinte top end.
        max_version = [INF] * num_digits

    if min_version > max_version:
        raise ValueError("The mininum version *MUST* be less than the maximum version.")

    # The very first version strategy we can have, is simply matching whatever the
    # mininum version is.
    version_strategies = [st.tuples(*[st.just(i) for i in min_version])]

    # Now we have to build up a list of possible versions besides our basic one.
    while min_version:
        # We're going to start with incrementing the rightmost digit in our version.
        incrementing_part = min_version.pop()

        # If the number of digits we would require to handle a version that is
        # larger than this mininum version is greater than the number of digits
        # we're trying to generate in a version, then we'll skip it and move onto
        # the next one.
        # Note: We add one to this to account for the incrementing_part that we removed
        #       from this list earlier.
        if len(min_version) + 1 > num_digits:
            continue

        # We're going to be making a version that has the same prefix as min_version,
        # but the incrementing part is one higher. If doing that would make the version
        # number we're just about to generate greater than our maximum version, then
        # we'll break out of the loop. Any further incrementing will continue to be
        # too large of a version number.
        if min_version + [incrementing_part + 1] > max_version[: len(min_version) + 1]:
            break

        # We're going to limit our generated version by the right most digit in our
        # maximum version.
        max_incrementing_part = max_version[len(min_version)]

        # Build up a parts that is all of the preceding digits, sans the final
        # digit, e.g. if our minimum version is 1.5.6.0, then we want 1, 5, 6.
        # We know this is safe with the maximum version, becasue if it wasn't, then
        # we would have bailed out earlier.
        parts = [st.just(i) for i in min_version]

        # If there are any values where the incrementing part will *always* mean that
        # any version number we generate, no matter what gets generated for the padded
        # versions, then we'll create strategies to deal with those first.
        if min_version + [incrementing_part + 1] < max_version[: len(min_version) + 1]:
            # if incrementing_part + 1 < max_incrementing_part:
            if (
                max_incrementing_part is INF
                or min_version != max_version[: len(min_version)]
            ):
                max_incr_value = None
            else:
                max_incr_value = max_incrementing_part - 1
            subparts = [
                st.integers(min_value=incrementing_part + 1, max_value=max_incr_value)
            ]

            # At this part, we know we can just blindly generate any padding we want,
            # because our leading digits will ensure that we are *always* less than
            # our maximum version.
            # Note: We have to subtract an extra 1 from our number of needed parts to
            #       complete our padding, because of the one we generated above.
            subparts += [
                st.integers(min_value=0) for _ in range(num_digits - len(parts) - 1)
            ]

            # Now we're going to create a hypothesis tuple from our prefix parts, and
            # our subparts, and add it to our list of strategies to try.
            version_strategies.append(st.tuples(*parts + subparts))

        # Finally, we will generate a strategy that sets the incrementing part and all
        # padded parts maximum value to be equal to the maximum value for that part in
        # our maximum value. The only special case here is that Infinity values in our
        # maximum values need to be translated to None for hypothesis. We need one
        # special case here, if our max_incrementing_part is inf, then this case should
        # already have been handled up above.
        if (
            max_incrementing_part is not INF
            and min_version == max_version[: len(min_version)]
        ):
            parts += [st.just(max_incrementing_part)]

            parts += [
                st.integers(min_value=0, max_value=_none_for_inf(max_version[i]))
                for i in range(len(parts), num_digits)
            ]

            # Create a hypothesis tuple from our parts, and add it to our list of
            # strategies to try.
            version_strategies.append(st.tuples(*parts))

    version = draw(st.one_of(version_strategies))

    # Now that we have a list of version strategies, we'll draw from one of those
    # possible strategies, and join the parts together to create a verison number.
    return ".".join(str(i) for i in version)
